/*********
https://www.baldengineer.com/arduino-lcd-display-tips.html
https://www.makerguides.com/character-i2c-lcd-arduino-tutorial/
https://deepbluembedded.com/lcd-custom-character-generator/
********
// https://circuitdigest.com/microcontroller-projects/littlefs-with-esp8266-to-read-write-and-delete-data-on-flash-memory-of-nodemcu
Arduino example code to display custom characters on I2C character LCD. More info:
www.www.makerguides.com
*/
//
#include "LiquidCrystal_I2C.h"
#include <LittleFS.h>

char FileName[] = __FILE__;
const char compile_date[] = __DATE__ " " __TIME__;

/////////////////////////////////////////////////////////////////////////////////////
int CUnits = 60 * 1000  ;   //multiplier for millis
//multiplier for timer  1 x1000 -> seconds, 60 -> minutes
//int CUnits = 60*60*1000  ; // hours  multiplier of millis()
/////////////////////////////////////////////////////////////////////////////////////

// Create lcd object of class LiquidCrystal_I2C:
LiquidCrystal_I2C lcd = LiquidCrystal_I2C(0x3f, 16, 2); // Change to (0x27,20,4) for 20x4 LCD.
// LCD lines
//chrars + 1=17
//char line0[17];
//char line1[17];
// values T1 T2
char LCDline[17]; //  16 char + null

long int TiMs  ;       // ms val when timer exhaust
long int lastTiMs = 0; // when status On or Off , store last timer  no more used
long int OnTiMs ;

//long int Tip = 0;  // ms last Ti
int T1 ;  // value preset
int T2 ;
int Tx ;  // last used timer nr, 1 or 2  set at power on
// default timer values T1 T2 see  CUnits
int T1d =  30 ;  //def val
int T2d =  60 ;
int Txd =   1 ;  //def to timer 1

// status  on   off  tmr
//         1    0    2
char statusTxt[6]; //"timer" 5 + null
int status = 2 ;  //def timer

/////////////////////////////////////////////////////////////////////////////////////
// Make custom characters:
// char is cell  largeur x hauteur = 5 x 8  pixels    On OFF Timer
//uint8_t T10[] = {0x17, 0x15, 0x17, 0x00, 0x1d, 0x08, 0x09, 0x09}; //i o T

// file storing Tx
const char* filename = "/Tx.txt";
/////////////////////////////////////////////////////////////////////////////////////
// set pin numbers
const int buttonT1 = 0;     // the GPIO# of the pushbutton pin
const int buttonT2 = 2;     // the GPIO# of the pushbutton pin
const int buttonPl = 14 ;   // the GPIO# of the pushbutton pin
const int buttonMi = 12;    // the GPIO# of the pushbutton pin
const int buttonOn = 13;    // the GPIO# of the pushbutton pin

const int ledPin =  16 ;    // the GPIO# of the LED pin

#define LEDON   HIGH
#define LEDOFF  LOW

/////////////////////////////////////////////////////////////////////////////////////
int writeT() {
  // Écrire les valeurs de T1 et T2 dans le fichier
  File file = LittleFS.open(filename, "w");
  if (file) {
    file.println(T1, 0); // (value, decimal places)
    file.println(T2, 0);
    file.println(Tx, 0);
    file.close();
    Serial.println("  writeT* écrits dans le fichier");
    Serial.printf ("  T1:%i T2:%i Tx:%i", T1, T1, Tx);
    return  true ;
  } else {
    Serial.println("  Erreur d'ouverture du fichier pour écriture");
    return false ;
  }
  Serial.println("") ;
}
/////////////////////////////////////////////////////////////////////////////////////
void waitAfterTimer() { //called at timer exhaust
  Serial.println("  waitAfterTimer()");
  int wl = 1 ;
  char LCDlineAlt[17];
  char waittxt[17] = "TmrOut,waiting.." ; //will alttext with time
  //                  1234567890123456
  int waittxtdelay = 700 ; //alt text delay
  long int MsZero = millis();

  sprintf(LCDline, "         Off    ");  // status or what else ?..
  //                1234567890123456
  lcd.setCursor(0, 1); // second line
  lcd.print(LCDline);

  sprintf(LCDlineAlt, "%s", waittxt );
  long int waitTgMs = millis() ;
  //  swap between two lines
  bool waitTg = 0 ;
  while (wl) {//loop until button
    yield() ;
    if (!waitTg) {
      sprintf(LCDline, "Wait%s %s", convertMilsToHHMMSS(millis() - MsZero), statusTxt );
      //                waitHH:MM:SS Off
      //                1234567890123456
      lcd.setCursor(0, 0);
      lcd.print(LCDline);
    }
    if (!digitalRead(buttonOn)||!digitalRead(buttonT1)||!digitalRead(buttonT2) )  {
      //loop button
      //pushed ..
      yield() ;
      lcd.setCursor(0, 0);
      lcd.print(LCDline);
      Serial.println("buttonOn");
      wl = 0 ;
    }  else {
      yield() ;
      if (millis() > waitTgMs + waittxtdelay ) { //after x sec ..
        lcd.setCursor(0, 0);
        lcd.print(LCDlineAlt);
        waitTgMs = millis() ;
        waitTg = !waitTg  ;
      }
    }
  }
}
/////////////////////////////////////////////////////////////////////////////////////
void setTiMs(int value = 0) {
  // value=0, get used timer Tx, or add/dec with passed value
  // return new TiMs
  long int tmil = millis() ;
  long int ret ;
  //Serial.println() ;
  Serial.printf(">  setTiMs() : %i", value ); Serial.println() ;
  //Serial.printf("  setTiMs() : cur TiMs %i",TiMs ); Serial.println() ;

  if (value == 0) {
    Serial.println("> value==0") ;
    long int TiOffset ;
    if (Tx == 1) {
      TiOffset = T1 * CUnits ;
    } else {
      TiOffset = T2 * CUnits ;
    }
    Serial.printf("> TiOffset : %i", TiOffset );   Serial.println() ;
    tmil = millis() ;
    TiMs = tmil + TiOffset  ;
  }
  else {
    Serial.printf("> setTiMs() : %i", value ); Serial.println() ;
    //TiMs= TiMs + value ;
    TiMs = tmil + (value * CUnits )  ;   //      TiMs=  millis() + (T1 *CUnits )  ;
    lastTiMs = TiMs ;
  }
  //Serial.printf("  millis() : %i", tmil );
  Serial.printf("> millis() : %i", tmil );   Serial.println() ;
  Serial.printf("> new TiMs : %i", TiMs );   Serial.println() ;
}
/////////////////////////////////////////////////////////////////////////////////////
String convertMilsToHHMMSS(long int mils) {  // millis to  HH:MM:SS
  char ret[9];
  long int seconds = mils / 1000;
  long int hours = seconds / 3600;
  seconds %= 3600;
  long int minutes = seconds / 60;
  seconds %= 60;
  long int secs = seconds;
  sprintf(ret, "%02d:%02d:%02d", hours, minutes, seconds) ; //8 chars
  return ret ;
}
/////////////////////////////////////////////////////////////////////////////////////
void handlestatus() {
  switch (status) {
    case 0: // OFF
      sprintf(statusTxt, "%s", String("Off  " ));  //5 chars
      //                               12345
      OnTiMs = millis();
      //lastTiMs=TiMs ;
      digitalWrite(ledPin, LEDOFF);
      break ;
    case 1: // ON/OFF  and count duration
      sprintf(statusTxt, "%s", String("On   ") );
      OnTiMs = millis();
      //lastTiMs=TiMs ;
      digitalWrite(ledPin, LEDON);
      break ;
    case 2: // timer
      sprintf(statusTxt, "%s", String( "Timer") );
      //TiMs= lastTiMs ;
      setTiMs() ;
      lastTiMs = 0 ;
      break ;
  }
}
/////////////////////////////////////////////////////////////////////////////////////
void setup() {
  Serial.begin(115200);  // start serial port
  delay(500);
  Serial.println("");
  Serial.println("");
  Serial.println("━━━━━━━━━━━━━━━━━━━━━━━━━━━━━");
  //              Compiled Nov  6 2023 19:00:57
  Serial.println(__FILE__);
  Serial.println(String("Compiled ") + String(compile_date) ) ;
  Serial.println("━━━━━━━━━━━━━━━━━━━━━━━━━━━━━");
  //////////////////////////////////////////
  Serial.println("Read LittleFS Tx values:");
  //////////////////////////////////////////
  if (!LittleFS.begin()) {
    Serial.println("  Erreur de montage LittleFS");
    return;
  }
  Serial.println(" Read stored Tx:");
  if (LittleFS.exists(filename)) {
    File file = LittleFS.open(filename, "r");
    if (file) {
      String line = file.readStringUntil('\n');
      T1 = line.toInt();
      line = file.readStringUntil('\n');
      T2 = line.toInt();
      line = file.readStringUntil('\n');
      Tx = line.toInt();
      file.close();
    } else {
      Serial.println("  Erreur d'ouverture du fichier pour lecture");
    }
  } else {
    T1 = T1d ;
    T2 = T2d ;
    Tx =  Txd ;
    Serial.println("  > File not found, set default values.");
    if (writeT() ) {
      Serial.println(" ok, writeT()has stored data");
    }
  }
  Serial.print("  T1 : "); Serial.println(T1);
  Serial.print("  T2 : "); Serial.println(T2);
  Serial.print("  Tx : "); Serial.println(Tx);
  //////////////////////////////////////////
  Serial.println(F("Initialize LCD.."));
  //////////////////////////////////////////
  lcd.init();
  lcd.backlight();  //  noBacklight() ;
  //lcd.createChar(1, T10 ) ;  //not used
  // Clear the LCD screen:
  lcd.clear();
  // Print a message to the lcd:
  //         1234567890123456
  //lcd.print("CustomCharacter1");
  // will be erased by loop() except if crash :-)
  //lcd.print(__FILE__);

  //////////////////////////////////////////
  Serial.println(F("Set pins.."));
  //////////////////////////////////////////
  // initialize the pushbutton pin as an input
  pinMode(buttonT1, INPUT_PULLUP);
  pinMode(buttonT2, INPUT_PULLUP);
  //pinMode(buttonT3, INPUT_PULLUP);
  pinMode(buttonPl, INPUT_PULLUP);
  pinMode(buttonMi, INPUT_PULLUP);
  pinMode(buttonOn, INPUT_PULLUP);
  // initialize the LED pin as an output
  pinMode(ledPin, OUTPUT);
  // power on at start
  digitalWrite(ledPin, LEDON);
  //////////////////////////////////////////
  Serial.println(F("Initialize status, TiMs.."));
  //////////////////////////////////////////
  handlestatus();
  setTiMs();
  Serial.println(F("End Setup()"));
}
/////////////////////////////////////////////////////////////////////////////////////
void loop() {
  //To select a cursor location set the column 0..1 THEN the row 0..15
  //top left is 0,0
  // Like: lcd.setCursor(column, row)
  /*
    lcd.setCursor(0, 1);  // col 2, l 1
    lcd.write(1);  /ioT  char
  */
  // Timer exhausted  ?
  long int t = millis();   //time flies ..
  //TiMs check

  if (status == 2) { // when timer active
    if (t > TiMs) { // if exhausted
      status = 0 ; handlestatus();
      Serial.printf(" TiMS = %i  exhausted, wait buttonOn", TiMs )  ;
      //Serial.println(F(" exhausted, wait buttonOn"));
      digitalWrite(ledPin, LEDOFF);
      waitAfterTimer() ;  // loop until pushed
      digitalWrite(ledPin, LEDON);
      setTiMs (0) ;
      Serial.println(F(" timer reset"));
    }
  }
  //else {  // when status ON or OFF
  //  }
  ///////////////////////////////////////////////////////////////////
  //  show first line
  //int tpct ;  //  pct of time elapsed

  //String tpct="na pct" ;  //  pct of time elapsed   000pct
  String tpct="      " ;  //  pct of time elapsed   000pct
  int T ; //hold on value for T1 T2

  if (status==2) { //pct only if timer on  //
    if (Tx == 1) { T = T1; } else  { T = T2; }
    tpct = String((TiMs-t)*100/(T*CUnits))+" pct  " ;  // "  " to avoid double final't' when reaching <=9 pct, cast to 6 on sprintf
  } else {
    //tpct="";
  }
//
  if (Tx == 1)  { //  set visual indication of active timer value  >  <
    sprintf(LCDline, ">%-3i<%-3i  %6s", T1, T2, tpct) ;  }
  else {
    sprintf(LCDline, " %-3i>%-3i< %6s", T1, T2, tpct) ;  }
  //                 "1234567890123456"
  //                  >123<    .
  //                      >123< xxxpct
  //                      >123<
  lcd.setCursor(0, 0);
  lcd.print(LCDline);
  ///////////////////////////////////////////////////////////////////
  //  show second line
  String HHMMSS;
  if (status == 2) { //show timer active
    HHMMSS =  convertMilsToHHMMSS(TiMs - t) ;
  } else { // show On or Off duration
    HHMMSS =  convertMilsToHHMMSS(t - OnTiMs) ;
  }
  sprintf(LCDline, "%s   %s", HHMMSS, statusTxt );
  //                  123
  //                1234567890123456
  //                hh:mm:ss...Timer
  //                        123

  lcd.setCursor(0, 1);// show second line
  lcd.print(LCDline);

  ///////////////////////////////////////////////////////////////////
  // handle buttons ///////////////////////////////////////////////////////
  ///////////////////////////   T1
  if (!digitalRead(buttonT1) )  {
    Serial.println("buttonT1");
    delay(300);
    status = 2 ; handlestatus() ;
    Tx = 1 ;
    setTiMs(T1) ;
  }
  ///////////////////////////   T2
  if (!digitalRead(buttonT2) )  {
    Serial.println("buttonT2");
    delay(300);
    status = 2 ; handlestatus() ;
    Tx = 2 ;
    setTiMs(T2) ;
  }
  ///////////////////////////   + Plus
  if (!digitalRead(buttonPl) )  {
    Serial.println("buttonPl");
    delay(300);
    if (Tx == 1) {
      T1++ ;  // no limits  live dangerously :-)
    }  else {
      T2++ ;
    }
    setTiMs(0);
    writeT() ;
  }
  ///////////////////////////   - Minus
  if (!digitalRead(buttonMi) ) {
    Serial.println("buttonMi");
    delay(300);
    if (Tx == 1) {
      T1-- ;
      if (T1 == 0) {
        T1 = 1;
      }
    }
    else {
      T2-- ;
      if (T2 == 0) {
        T2 = 1;
      }
    }
    setTiMs(0);
    writeT() ;
  }
  ///////////////////////////   OnOffTmr
  if (!digitalRead(buttonOn) )  {
    Serial.println("buttonOn");
    Serial.printf(" -- status:%i  TiMs=%i  \n", status , TiMs  ) ;
    Serial.printf(" - Tx:%i /T1:%i /T2:%i /timer:%i", Tx, T1, T2, TiMs / CUnits); Serial.println();

    delay(300);
    status++; handlestatus();
    if (status == 3) {
      status = 0 ;
    }
    if (status == 2) { // reset timer to Tx val
      Serial.println(">Reset Timer");
      setTiMs(0) ;
    }
    handlestatus() ;  //because ..  why not
  }
}
/////////////////////////////////////////////////////////////////////////////////////
